"""This module contains the load estimator of the grid operator."""

from midas.modules.dlpdata.simulator import DLPSimulator


class LoadEstimator:
    """The load estimator class of the grid operator."""

    def __init__(self, params, entity_map):
        self.config = params
        self.entity_map = entity_map

        sim = DLPSimulator()
        sim.init(
            "",
            step_size=1,
            interpolate=True,
            seed=0,
            start_date=self.config["start_date"],
            data_path=self.config["data_path"],
        )
        self.models = dict()

        # Create estimators for household models
        # for bus_id, loads in self.config['land_mapping'].items():
        #     self.models.setdefault(bus_id, list())
        #     for (_, p_mwh_per_a) in loads:
        #         entity = sim.create(1, 'DefaultLoadH0',
        #                             p_mwh_per_a=p_mwh_per_a*0.1)[0]
        #         self.models[bus_id].append(sim.models[entity['eid']])

        for bus_id, loads in self.config["load_mapping"].items():
            self.models.setdefault(bus_id, list())
            for (_, p_mwh_per_a) in loads:
                entity = sim.create(1, "H0", p_mwh_per_a=p_mwh_per_a)[0]
                self.models[bus_id].append(sim.models[entity["eid"]])

        self.now_dt = None
        self.setpoints = None

    def step(self):
        """Perform a simulation step."""

        self.setpoints = {"Load": dict()}
        for bus_id, models in self.models.items():
            for info in self.entity_map.values():
                if info["etype"] != "Load":
                    continue
                static = info["static"]
                if static["bus"] != bus_id:
                    continue
                idx = info["idx"]

                p_mws = list()
                q_mvars = list()

                for model in models:
                    model.now_dt = self.now_dt
                    model.cos_phi = 0.9
                    model.step()
                    p_mws.append(model.p_mw)
                    q_mvars.append(model.q_mvar)

                self.setpoints["Load"].setdefault(
                    idx,
                    {
                        "static": static,
                        "attrs": {"p_mw": sum(p_mws), "q_mvar": sum(q_mvars)},
                    },
                )



"""This module contains the observer of the grid operator."""


class GridObserver:
    """The observer model.

    Runs the powerflow with currently known information.

    """

    def __init__(self, params, grid):

        self.model = grid
        self.config = params
        self.setpoints = None
        self.estimations = None

    def step(self):
        """Perform a simulation step."""

        if self.estimations is not None:
            self._assign_values(self.estimations)

        self._assign_values(self.setpoints)

        self.model.run_powerflow(0)

        self.setpoints = None
        self.estimations = None

    def _assign_values(self, setpoints):

        for etype, idxs in setpoints.items():
            for idx, attrs in idxs.items():
                self.model.set_inputs(etype, idx, attrs["attrs"])



"""This module contains the forecaster of the grid operator."""
from datetime import timedelta
import numpy as np
import pandas as pd
from sortedcontainers import SortedDict

from .grid_observer import GridObserver
from .grid_analyzer import GridAnalyzer


class GridForecaster:
    """The grid forecaster calculates future grid states."""

    def __init__(self, grid, params, load_estimator):
        self.config = params
        self.observer = GridObserver(self.config, grid)
        self.analyzer = GridAnalyzer(grid, params)
        self.load_estimator = load_estimator
        self.outbox = None

        self.now_dt = None
        self.step_size = 0
        self.forecasts = None
        self._setpoints = None
        self.actions = None

    def step(self):
        """Perform a simulation step."""
        self.actions = dict()
        fc_dt = pd.date_range(
            self.now_dt + timedelta(seconds=self.step_size),
            periods=self.config["forecast_horizon_hours"]
            * 3_600
            / self.step_size,
            freq="{}S".format(self.step_size),
        )
        for date in fc_dt:
            self._run_forecast(date)
            self._recommend_actions(self.analyzer.stats, date)

        self._send_schedules()

    def _run_forecast(self, date):
        """Run the forecast for one step (*date*)."""
        setpoints = SortedDict()
        for etype, idxs in self.forecasts.items():
            setpoints.setdefault(etype, SortedDict())

            for idx, src_ids in idxs.items():
                setpoints[etype].setdefault(idx, SortedDict())
                setpoints[etype][idx].setdefault("static", src_ids["static"])
                setpoints[etype][idx].setdefault("attrs", SortedDict())

                self.load_estimator.now_dt = date
                self.load_estimator.step()

                for src_id, schedule in src_ids["src"].items():
                    if not isinstance(schedule, pd.DataFrame):
                        schedule = pd.read_json(schedule)
                        src_ids["src"][src_id] = schedule
                    for name in ["p_mw", "q_mvar"]:
                        setpoints[etype][idx]["attrs"].setdefault(name, 0)
                        val = float(schedule.loc[date][name])
                        if etype == "Sgen" and "Pysimmod" in src_id:
                            # Transform to load reference system
                            val *= -1
                        setpoints[etype][idx]["attrs"][name] += val

        # self.observer.estimations = self.load_estimator.setpoints
        self.observer.setpoints = setpoints
        self.observer.step()

        self.analyzer.eval_voltage_violations()

    def _recommend_actions(self, stats, date):
        """Analyze the results of the grid analyzer

        Currently, only the over voltage case is handled.

        """
        if stats["voltage_violations"]["num_violations"] == 0:
            # No action required
            return

        violation_idxs = np.where(
            stats["voltage_violations"]["upper_violations"] > 0.0
        )[0]

        for bus_idx in violation_idxs:
            for etype, idxs in self.forecasts.items():
                if etype in ["Load", "Sgen"]:
                    for _, srcs in idxs.items():

                        self._create_setpoints(
                            srcs["src"], date, bus_idx, more_power=False
                        )
                else:
                    # Handle tranformers, switch, etc
                    pass

        violation_idxs = np.where(
            stats["voltage_violations"]["lower_violations"] < 0.0
        )[0]

        for bus_idx in violation_idxs:
            for etype, idxs in self.forecasts.items():
                if etype in ["Load", "Sgen"]:
                    for _, srcs in idxs.items():

                        self._create_setpoints(
                            srcs["src"], date, bus_idx, more_power=True
                        )
                else:
                    # Handle tranformers, switch, etc
                    pass

    def _create_setpoints(self, srcs, date, bus_idx, more_power):
        """Fill the actions dictionary will schedules."""

        for src_id, schedule in srcs.items():
            entity_bus = self.config["der_mapping"][src_id]["bus"]
            if entity_bus != bus_idx:
                continue

            if self.config["der_mapping"][src_id]["type"] == "load":
                if more_power:
                    if schedule.loc[date]["target"] > 0.5:
                        setpoint = 0.5
                    elif schedule.loc[date]["target"] > 0.0:
                        setpoint = 0.0
                    else:
                        continue
                else:
                    if schedule.loc[date]["target"] < 0.5:
                        setpoint = 0.5
                    elif schedule.loc[date]["target"] < 1.0:
                        setpoint = 1.0
                    else:
                        continue

            elif self.config["der_mapping"][src_id]["type"] == "sgen":
                if more_power:
                    if schedule.loc[date]["target"] < 0.5:
                        setpoint = 0.5
                    elif schedule.loc[date]["target"] < 1.0:
                        setpoint = 1.0
                    else:
                        continue
                else:
                    if schedule.loc[date]["target"] > 0.5:
                        setpoint = 0.5
                    elif schedule.loc[date]["target"] > 0.0:
                        setpoint = 0.0
                    else:
                        continue
            else:
                continue
            self.actions.setdefault(src_id, dict())
            self.actions[src_id][date] = setpoint

    def _send_schedules(self):
        """Unpack the actions dictionary an create messages."""
        for dst_id, schedule in self.actions.items():
            self.outbox.send(
                {
                    "to": dst_id,
                    "topic": "schedule",
                    "msg": pd.DataFrame(
                        data=list(schedule.values()),
                        columns=["target"],
                        index=list(schedule.keys()),
                    ),
                }
            )



"""This module contains the grid analyzer component."""
import numpy as np


class GridAnalyzer:
    """Grid analyzer of the grid operator."""

    def __init__(self, grid, config):
        self.model = grid
        self.config = config
        self.real_grid_data = dict()
        self.stats = dict()

    def step(self):
        """Perform a simulation step."""

        self._eval_grid_state()

    def _eval_grid_state(self):
        """Calculate different metrics describing the grid state.

        Currently, the average voltage p.u. is calculated.
        """
        self.stats = dict()
        self.eval_observation_error()
        self.eval_voltage_violations()

        self.stats["health"] = float(
            self.model.grid.res_bus.vm_pu.values[1:].mean()
        )

    def eval_observation_error(self):
        """Evaluate the observation error if feasible."""

        if not self.real_grid_data:
            # We don't have grid data
            self.stats["error"] = 0
            return

        if "Bus" in self.real_grid_data:
            # Move to function 'get_real_data'
            real = dict()
            for bus_id, attrs in self.real_grid_data["Bus"].items():
                # if bus_id == 0:
                #     # We skip the slack bus
                #     continue
                if self.config["grid_busvm_sensor"]:
                    real.setdefault("vm_pu", list()).append(attrs["vm_pu"])
                elif self.config["grid_busload_sensor"]:
                    real.setdefault("p_mw", list()).append(attrs["p_mw"])
                    real.setdefault("q_mvar", list()).append(attrs["q_mvar"])

            if self.config["grid_busvm_sensor"]:
                real["vm_pu"] = np.array(real["vm_pu"]).flatten()
                obs = np.array(
                    list(self.model.grid.res_bus.vm_pu.values[1:])
                ).flatten()
                self.stats["bus.vm_pu"] = {
                    "errors": real["vm_pu"] - obs,
                    "rmse": rmse(real["vm_pu"], obs),
                }
                self.stats["error"] = float(self.stats["bus.vm_pu"]["rmse"])

            elif self.config["grid_busload_sensor"]:
                real["p_mw"] = np.array(real["p_mw"]).flatten()
                real["q_mvar"] = np.array(real["q_mvar"]).flatten()
                obs = {
                    "p_mw": np.array(
                        list(self.model.grid.res_bus.p_mw.values[1:])
                    ).flatten(),
                    "q_mvar": np.array(
                        list(self.model.grid.res_bus.q_mvar.values[1:])
                    ).flatten(),
                }
                self.stats["bus.p_mw"] = {
                    "errors": real["p_mw"] - obs["p_mw"],
                    "rmse": rmse(real["p_mw"], obs["p_mw"]),
                }
                self.stats["bus.q_mvar"] = {
                    "errors": real["q_mvar"] - obs["q_mvar"],
                    "rmse": rmse(real["q_mvar"], obs["q_mvar"]),
                }
                self.stats["error"] = float(self.stats["bus.p_mw"]["rmse"])

            else:
                # Handle grid data other than buses
                self.stats["error"] = 0

    def eval_voltage_violations(self):
        """Evaluate possible voltage violations."""
        num_buses = len(self.model.grid.res_bus)

        lower_bounds = np.ones(num_buses) * self.config["undervoltage_pu"]
        upper_bounds = np.ones(num_buses) * self.config["overvoltage_pu"]

        lower_violations = self.model.grid.res_bus.vm_pu - lower_bounds
        upper_violations = self.model.grid.res_bus.vm_pu - upper_bounds

        self.stats["voltage_violations"] = {
            "lower_violations": lower_violations.values,
            "upper_violations": upper_violations.values,
            "max_overvoltage": float(max(0, upper_violations.max())),
            "min_undervoltage": float(min(0, lower_violations.min())),
            "num_violations": int(
                np.where(lower_violations < 0, 1, 0).sum()
                + np.where(upper_violations > 0, 1, 0).sum()
            ),
        }


def rmse(real, pred):
    """Compute the root mean squared error."""
    return ((real - pred) ** 2).sum() ** 0.5



"""This module contains the coordinator for the subfunctions of
the grid operator.

"""
from datetime import datetime, timedelta

from midas.modules.powergrid.model.static import PandapowerGrid

from .grid_analyzer import GridAnalyzer
from .grid_forecaster import GridForecaster
from .grid_observer import GridObserver
from .load_estimator import LoadEstimator
from .messenger import Messenger


class Coordinator:
    """Coordinator for all the subfunctions of the grid operator."""

    def __init__(self, params):

        self.config = params

        self.ppg = PandapowerGrid()
        self.ppg.setup(self.config["gridfile"], 0)

        self.now_dt = datetime.strptime(
            self.config["start_date"], "%Y-%m-%d %H:%M:%S%z"
        )
        self.step_size = self.config.get("step_size", 900)

        self.messenger = Messenger(self.config, self.ppg.entity_map)
        self.observer = GridObserver(self.config, self.ppg)
        self.analyzer = GridAnalyzer(self.ppg, self.config)
        self.load_estimator = LoadEstimator(self.config, self.ppg.entity_map)
        self.forecaster = GridForecaster(
            self.ppg, self.config, self.load_estimator
        )
        self.forecaster.outbox = self.messenger

        self.error = None
        self.health = None
        self.min_undervoltage = None
        self.max_overvoltage = None
        self.num_voltage_violations = None

    def step(self):
        """Perform a simulation step."""

        # Read inbox and filter messages
        self.messenger.step()

        if not self.config["grid_load_sensor"]:
            # Run the load estimator
            self.load_estimator.now_dt = self.now_dt
            self.load_estimator.step()
            self.observer.estimations = self.load_estimator.setpoints

        # Pass setpoints to grid observer
        self.observer.setpoints = self.messenger.setpoints
        self.observer.step()

        # Run the analyzer
        self.analyzer.real_grid_data = self.messenger.real_grid_data
        self.analyzer.step()

        self.error = self.analyzer.stats["error"]
        self.health = self.analyzer.stats["health"]
        self.min_undervoltage = self.analyzer.stats["voltage_violations"][
            "min_undervoltage"
        ]
        self.max_overvoltage = self.analyzer.stats["voltage_violations"][
            "max_overvoltage"
        ]
        self.num_voltage_violations = self.analyzer.stats[
            "voltage_violations"
        ]["num_violations"]

        if self.config["run_forecast"]:
            # Forecast the next steps
            self.forecaster.forecasts = self.messenger.forecasts
            self.forecaster.now_dt = self.now_dt
            self.forecaster.step_size = self.step_size
            self.forecaster.step()

        self.now_dt += timedelta(seconds=self.step_size)

    def receive(self, msg):
        """Receive a message.

        The message is a dictionary containing three keys:
        'from': (id of the source, e.g. '<sid>.<eid>'),
        'topic': (context of the message, e.g., the attribute to set),
        'msg': (content of the message, e.g., value to set)

        """
        self.messenger.inbox.put(msg)

    @property
    def outbox(self):
        """The message outbox."""
        return self.messenger.outbox

    @property
    def grid(self):
        return self.ppg.to_json()



"""This module contains the Messenger, which handles
the communication stuff of the grid operator.

"""
from queue import Queue
from sortedcontainers import SortedDict


class Messenger:
    """Handles incoming and outgoing messages."""

    def __init__(self, params, entity_map):
        self.config = params
        self.entity_map = entity_map
        self.inbox = Queue()
        self.outbox = Queue()

        self.setpoints = None
        self.forecasts = None
        self.real_grid_data = None

    def step(self):
        """Perform a simulation step."""

        # Read the inbox
        messages = self._read_inbox()

        # Filter the messages
        self._filter_messages(messages)

    def receive(self, msg):
        """Receive a message.

        The message is a dictionary containing three keys:
        'from': (id of the source, e.g. '<sid>.<eid>'),
        'topic': (context of the message, e.g., the attribute to set),
        'msg': (content of the message, e.g., value to set)

        """
        self.inbox.put(msg)

    def send(self, msg):
        """Send a message.

        The message is a dictionary containing three keys:
        'to': (id of the destination, e.g., '<sid>.<eid>'),
        'topic': (context of the message, e.g., the attribute to set),
        'msg': (content of the message).

        """
        self.outbox.put(msg)

    def _read_inbox(self):
        """Read all messages from the inbox and wrap them into a
        dictionary.

        """
        messages = SortedDict()
        while not self.inbox.empty():
            msg = self.inbox.get()
            # msg_to = self.gridcfg['connections']['ingoing'][msg['from']]
            if msg["from"] not in messages:
                messages[msg["from"]] = {msg["topic"]: list()}
            if msg["topic"] not in messages[msg["from"]]:
                messages[msg["from"]][msg["topic"]] = list()
            messages[msg["from"]][msg["topic"]].append(msg["msg"])
        return messages

    def _filter_messages(self, messages):
        """Filter the messages according to their purpose."""
        self.setpoints = SortedDict()
        self.forecasts = SortedDict()
        self.real_grid_data = SortedDict()

        for dst, attrs in messages.items():
            _, eid = dst.split(".")[:2]
            if eid in self.entity_map:
                # print('In EntityMap', dst, attrs)
                self._handle_entity_map(eid, attrs)
            elif dst in self.config["der_mapping"]:
                self._handle_unit_map(dst, attrs)
                # print('In UnitMapping', dst, attrs)
            else:
                # print('Without category', dst, attrs)
                pass

    def _handle_entity_map(self, eid, attrs):
        """Handle messages with destinations contained in the
        entity map.

        """
        idx = self.entity_map[eid]["idx"]
        etype = self.entity_map[eid]["etype"]
        static = self.entity_map[eid]["static"]

        if etype.lower() in ["load", "sgen"]:
            self.setpoints.setdefault(etype, SortedDict()).setdefault(
                idx, SortedDict()
            )
            self.setpoints[etype][idx].setdefault("attrs", SortedDict())
            self.setpoints[etype][idx].setdefault("static", static)
            for attr, vals in attrs.items():
                self.setpoints[etype][idx]["attrs"].setdefault(attr, 0)
                self.setpoints[etype][idx]["attrs"][attr] += float(sum(vals))

        elif etype.lower() in ["bus"]:
            self.real_grid_data.setdefault(etype, SortedDict())
            self.real_grid_data[etype].setdefault(idx, SortedDict())
            for attr, vals in attrs.items():
                self.real_grid_data[etype][idx][attr] = vals[0]

        else:
            # Handle transformers, lines, switches, etc
            pass

    def _handle_unit_map(self, dst, attrs):
        """Handle messages from unit models with grid as
        destination.

        """
        cfg = self.config["der_mapping"][dst]
        etype = cfg["type"]
        etype_cap = etype.capitalize()
        bidx = cfg["bus"]
        idx = 0
        static = None
        # for idx in range(len(getattr(self.model.gri)))
        for _, eattrs in self.entity_map.items():
            if (
                eattrs["etype"] == etype_cap
                and eattrs["static"]["bus"] == bidx
            ):
                idx = eattrs["idx"]
                static = eattrs["static"]

        self.setpoints.setdefault(etype_cap, SortedDict())
        self.setpoints[etype_cap].setdefault(idx, SortedDict())
        self.setpoints[etype_cap][idx].setdefault("attrs", SortedDict())
        self.setpoints[etype_cap][idx].setdefault("static", static)

        self.forecasts.setdefault(etype_cap, SortedDict())
        self.forecasts[etype_cap].setdefault(idx, SortedDict())
        self.forecasts[etype_cap][idx].setdefault("static", static)
        self.forecasts[etype_cap][idx].setdefault("src", SortedDict())

        for attr, vals in attrs.items():
            if attr in ["p_mw", "q_mvar"]:
                if is_input_from_unitmod(etype, self.config):
                    val = self.setpoints[etype_cap][idx]["attrs"].get(attr, 0)
                    val += float(sum(vals))
                    self.setpoints[etype_cap][idx]["attrs"][attr] = val
            elif attr == "schedule":
                # Should only be one schedule for each unit model!
                self.forecasts[etype_cap][idx]["src"][dst] = vals[0]


def is_input_from_unitmod(etype, flags):
    """Check whether the given etype should be considered as input."""
    if etype == "load" and not flags["grid_load_sensor"]:
        return True
    if etype == "sgen" and not flags["grid_sgen_sensor"]:
        return True
    return False



"""This module contains the MIDAS upgrade for the grid operator
simulator.
"""
import logging

from midas.util.upgrade_module import UpgradeModule

LOG = logging.getLogger(__name__)


class GridOperatorModule(UpgradeModule):
    """Grid operator upgrade module for MIDAS."""

    def __init__(self):
        super().__init__(
            module_name="goa",
            default_scope_name="midasmv",
            default_sim_config_name="GridOperator",
            default_import_str=(
                "midas.modules.goa.simulator:GridOperatorSimulator"
            ),
            default_cmd_str=(
                "%(python)s -m midas.modules.goa.simulator %(addr)s"
            ),
            log=LOG,
        )

        self.models = {
            "GOA": {
                "grid_load_sensor": ("load", ["p_mw", "q_mvar"]),
                "grid_sgen_sensor": ("sgen", ["p_mw", "q_mvar"]),
                "grid_busvm_sensor": ("bus", ["vm_pu"]),
                "grid_busload_sensor": ("bus", ["p_mw", "q_mvar"]),
                "der_load_sensor": ("load", ["p_mw", "q_mvar"]),
                "der_load_forecast_sensor": ("load", ["schedule"]),
                "der_sgen_sensor": ("sgen", ["p_mw", "q_mvar"]),
                "der_sgen_forecast_sensor": ("sgen", ["schedule"]),
                "goa_output": [
                    "health",
                    "error",
                    "max_overvoltage",
                    "min_undervoltage",
                    "num_voltage_violations",
                ],
            }
        }

    def check_module_params(self, module_params):
        """Check the module params and provide default values."""

        module_params.setdefault("start_date", self.scenario.base.start_date)
        module_params.setdefault("data_path", self.scenario.base.data_path)
        module_params.setdefault("grid_load_sensor", False)
        module_params.setdefault("grid_sgen_sensor", False)
        module_params.setdefault("grid_busvm_sensor", False)
        module_params.setdefault("grid_busload_sensor", True)
        module_params.setdefault("der_load_sensor", False)
        module_params.setdefault("der_load_forecast_sensor", False)
        module_params.setdefault("der_sgen_sensor", False)
        module_params.setdefault("der_sgen_forecast_sensor", False)
        module_params.setdefault("undervoltage_pu", 0.9)
        module_params.setdefault("overvoltage_pu", 1.1)
        module_params.setdefault("run_forecast", False)
        module_params.setdefault("forecast_horizon_hours", 0.25)

    def check_sim_params(self, module_params):
        """Check the params for a certain simulator instance."""

        self.sim_params.setdefault("start_date", module_params["start_date"])
        self.sim_params.setdefault("data_path", module_params["data_path"])
        self.sim_params.setdefault(
            "grid_load_sensor", module_params["grid_load_sensor"]
        )
        self.sim_params.setdefault(
            "grid_sgen_sensor", module_params["grid_sgen_sensor"]
        )
        self.sim_params.setdefault(
            "grid_busvm_sensor", module_params["grid_busvm_sensor"]
        )
        self.sim_params.setdefault(
            "grid_busload_sensor", module_params["grid_busload_sensor"]
        )
        self.sim_params.setdefault(
            "der_load_sensor", module_params["der_load_sensor"]
        )
        self.sim_params.setdefault(
            "der_load_forecast_sensor",
            module_params["der_load_forecast_sensor"],
        )
        self.sim_params.setdefault(
            "der_sgen_sensor", module_params["der_sgen_sensor"]
        )
        self.sim_params.setdefault(
            "der_sgen_forecast_sensor",
            module_params["der_sgen_forecast_sensor"],
        )
        self.sim_params.setdefault(
            "undervoltage_pu", module_params["undervoltage_pu"]
        )
        self.sim_params.setdefault(
            "overvoltage_pu", module_params["overvoltage_pu"]
        )
        self.sim_params.setdefault(
            "run_forecast", module_params["run_forecast"]
        )
        self.sim_params.setdefault(
            "forecast_horizon_hours", module_params["forecast_horizon_hours"]
        )
        self.sim_params.setdefault("gridfile", self.scope_name)

        self.gather_mappings()

        self.sim_params.setdefault("seed_max", self.scenario.base.seed_max)
        self.sim_params.setdefault("seed", self.scenario.create_seed())

    def start_models(self):
        """Start all models for this simulator."""

        for model in self.models:
            model_key = self.scenario.generate_model_key(self, model)
            self.start_model(model_key, model, {"params": self.sim_params})

    def connect(self):
        """Connect the models to existing other models."""
        for model, sensors in self.models.items():
            goa = self.scenario.generate_model_key(self, model)

            for sensor, info in sensors.items():
                if sensor.startswith("grid") and self.sim_params[sensor]:
                    self._connect_to_grid_entites(goa, info)
                elif sensor.startswith("der") and self.sim_params[sensor]:
                    self._connect_to_der_entities(goa, info)

    def _connect_to_grid_entites(self, goa, info):
        entities = self.scenario.find_grid_entities(
            self.sim_params["gridfile"], info[0]
        )
        for key in entities:
            self.connect_entities(key, goa, info[1])

    def _connect_to_der_entities(self, goa, info):
        models = self.scenario.find_models("der", self.scope_name)

        for key in models:
            self.connect_entities(key, goa, info[1])

    def connect_to_db(self):
        db_key = self.scenario.find_first_model("store", "database")[0]
        for model, sensors in self.models.items():
            attrs = sensors["goa_output"]
            goa = self.scenario.generate_model_key(self, model)
            self.connect_entities(goa, db_key, attrs)

    def get_sensors(self):
        for model, sensors in self.models.items():
            # attrs = sensors["goa_output"]
            goa = self.scenario.get_model(self.scenario.generate_model_key(self, model))
            self.scenario.sensors.append(
                {
                    "sensor_id": f"{goa.full_id}.health",
                    "observation_space": (
                        "Box(low=0, high=1.2, shape=(1,), dtype=np.float32)"
                    ),
                }
            )

            self.scenario.sensors.append(
                {
                    "sensor_id": f"{goa.full_id}.error",
                    "observation_space": (
                        "Box(low=0, high=2.0, shape=(1,), dtype=np.float32)"
                    ),
                }
            )

    def gather_mappings(self):
        relevant = ["sbdata", "sndata", "comdata", "der"]
        load_mapping = dict()
        sgen_mapping = dict()
        der_mapping = dict()
        shared_mappings = self.scenario.get_shared_mappings()

        for key, val in shared_mappings.items():
            for rel in relevant:
                if not key.startswith(rel):
                    continue
                elif self.scope_name not in key:
                    continue
                elif "eid_mapping" in key:
                    self._collect_eids(der_mapping, val)
                elif "load_mapping" in key:
                    self._collect_entities(load_mapping, val)
                elif "sgen_mapping" in key:
                    self._collect_entities(sgen_mapping, val)
        self.sim_params.setdefault("load_mapping", dict())
        if not self.sim_params["load_mapping"]:
            self.sim_params["load_mapping"] = load_mapping

        self.sim_params.setdefault("sgen_mapping", dict())
        if not self.sim_params["sgen_mapping"]:
            self.sim_params["sgen_mapping"] = sgen_mapping

        self.sim_params.setdefault("der_mapping", dict())
        if not self.sim_params["der_mapping"]:
            self.sim_params["der_mapping"] = der_mapping

    def _collect_entities(self, store_in, mapping):
        for bus, entities in mapping.items():
            store_in.setdefault(bus, list())
            store_in[bus].extend(entities)

    def _collect_eids(self, store_in, mapping):
        for eid, info in mapping.items():
            store_in[eid] = info



"""This module contains the mosaik_api simulator meta for the
grid operator simulator.
"""


META = {
    "type": "time-based",
    "models": {
        "GOA": {
            "public": True,
            "params": ["params"],
            "any_inputs": True,
            "attrs": [
                "inbox",  # send messages to GOA
                "outbox",  # receive messages from GOA
                "health",  # Grid state
                "error",  # Miss-estimation of the GOA
                "max_overvoltage",
                "min_undervoltage",
                "num_voltage_violations",
                "grid",
            ],
        }
    },
}



""" This module contains the GridOperator simulator."""
import logging
from typing import Any, Dict, List

import mosaik_api
from midas.util.dict_util import update
from midas.util.logging import set_and_init_logger

from .meta import META
from .model.coordinator import Coordinator as GOA

LOG = logging.getLogger(__name__)


class GridOperatorSimulator(mosaik_api.Simulator):
    """The GridOperator simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid: str = ""
        self.step_size: int = 0
        self.models: Dict[str, GOA] = {}
        self.cache: dict = {}

    def init(self, sid: str, **sim_params) -> Dict[str, Any]:
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).
        """
        self.sid = sid
        if "step_size" not in sim_params:
            LOG.debug(
                "Param *step_size* not provided. "
                "Using default step size of 900."
            )
        self.step_size = sim_params.get("step_size", 900)
        return self.meta

    def create(self, num: int, model: str, **model_params):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.
        """

        assert num == 1, "Only one operator at a time"

        eid = f"GOA_{len(self.models)}"
        self.models[eid] = GOA(params=model_params["params"])

        return [{"eid": eid, "type": "GOA"}]

    def step(
        self,
        time: int,
        inputs: Dict[str, Dict[str, Dict[str, Any]]],
        max_advance: int = 0,
    ):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        LOG.debug("At step %d received inputs %s", time, inputs)

        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():
                for src_id, val in src_ids.items():
                    if val is None:
                        continue
                    if attr == "inbox":
                        for msg in val:
                            self.models[eid].receive(msg)
                            LOG.debug("GOA %s received message %s", eid, msg)
                    else:
                        msg = {
                            "from": src_id,
                            "to": f"{self.sid}.{eid}",
                            "topic": attr,
                            "msg": val,
                        }
                        self.models[eid].receive(msg)
                        LOG.debug("GOA %s received message %s", eid, msg)

        for eid, goa in self.models.items():
            goa.step()

        return time + self.step_size

    def get_data(
        self, outputs: Dict[str, List[Any]]
    ) -> Dict[str, Dict[str, Any]]:
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = {}
        grid_json = {}

        for eid, attrs in outputs.items():
            data[eid] = {}
            goa = self.models[eid]

            for attr in attrs:
                if attr == "outbox":
                    outs = dict()
                    while not goa.outbox.empty():
                        msg = goa.outbox.get()
                        if msg["to"] not in outs:
                            outs[msg["to"]] = dict()
                        outs[msg["to"]][msg["topic"]] = msg["msg"]
                    data[eid][attr] = outs
                elif attr == "grid":
                    grid_json.setdefault(eid, {})["grid"] = goa.grid
                else:
                    # Currently, only health and error can be requested
                    data[eid][attr] = getattr(goa, attr)

        LOG.debug("Gathered outputs %s", data)
        update(data, grid_json)

        return data


if __name__ == "__main__":
    set_and_init_logger(0, "goa-logfile", "midas-goa.log", replace=True)
    LOG.info("Starting mosaik simulation...")
    mosaik_api.start_simulation(GridOperatorSimulator())
